/*
 * Inputs:
 *     Set<PluginSourceConfig> availableSourceConfigSet : the source configurations which should be scanned
 *     Map<String, String> triggerPropertyMap : parameters passed to the trigger invocation
 *     Map<String, String> requestPropertyMap: parameters to be placed upon the build request for matching source-configs
 *     Logger log : a logger instance
 *
 * Output:
 *     Set<PluginSourceConfig> matchingSourceConfigSet : an empty set used to return source configurations which should trigger
 *
 * Return:
 *     nothing
 */

import com.urbancode.anthill3.domain.property.PropertyValueGroupFactory;

String.metaClass.endsWithIgnoreCase = { String suffix ->
    int dl = delegate.length();
    int sl = suffix.length();
    return dl >= sl && delegate.substring(dl - sl, dl).equalsIgnoreCase(suffix);
};

// normalize URI-like strings
String.metaClass.normalize = {
    return delegate.replaceAll('(?<!:)//+', '/');
};

def parseLong = { value ->
   return value ? value.toLong() : null
}

def resolveValue = {def module, def value ->
    return ParameterResolver.resolveForSourceConfig(value, module.getContainer());
}

def resolveModuleProperty = {def module, def propertyName ->
    return resolveValue(module, module.getPropertyValue(propertyName)?.value);
}

def calculateRepositoryUrl = { String remoteBaseUrl, String remoteUrl ->
    def result = remoteUrl;
    if (remoteBaseUrl && !(remoteUrl =~ '[a-zA-Z]+://')) {
        result = remoteBaseUrl + '/' + remoteUrl;
    }
    return result.normalize();
}

def buildSourcePropertyMap = { module ->
    def result = [:]
    module.getPropertyMap().each{ key, value ->
        result[key] = value?.value
    }

    Long scRepoId = parseLong(result['repo']);
    if (scRepoId) {
        def pvgFactory = PropertyValueGroupFactory.getInstance();
        def repoPvg = pvgFactory.restore(scRepoId);
        result['repo'] = repoPvg.name;
        repoPvg.getPropertyMap().each{ key, value ->
            result['repo/'+key] = value?.value
        }
    }
    return result;
}

def getLastSlash = { 
    int lastSlash
    if (it.lastIndexOf('/') != -1) {
        lastSlash = it.lastIndexOf('/')
    }
    else if (it.lastIndexOf('\\') != -1) {
        lastSlash = it.lastIndexOf('\\')
    }
    else {
        lastSlash = it.length()-1
    }
    return lastSlash
}

def trimLeadingAndTrailingSlashes = {
    def result = it.replaceFirst('^[/\\\\]+', "").replaceFirst('[/\\\\]+$', "")
    return result
}

//
//
//

String revision = triggerPropertyMap['revision']
String code = triggerPropertyMap['code']
String dirsChanged = triggerPropertyMap['dirs-changed']
String filesChanged = triggerPropertyMap['files-changed']
String revisionDate = triggerPropertyMap['revision-date']
String repoPath = triggerPropertyMap['repo']
String repoCollection

if (dirsChanged) {
    int lastSlash = getLastSlash(dirsChanged)
    dirsChanged = dirsChanged.substring(0, lastSlash)
}
if (repoPath) {
    int lastSlash = getLastSlash(repoPath)
    repoCollection = repoPath.substring(lastSlash)
}

if (dirsChanged && revision) {

    for (sc in availableSourceConfigSet) {
        for (module in sc.getSourcePropertyValueGroups()) {
            def plugin = module.plugin
            if (plugin.pluginId.endsWith('.plugin.Subversion')) {
                Map<String,String> scProps = buildSourcePropertyMap(module)
                
                boolean isSvnRepo = scProps['repo/isSvnRepo']

                String scRepoBasePath = scProps['repo/repositoryUrl']
                String scRepoPath = resolveModuleProperty(module, "projectTrunkPath")

                String finalRepoPath = calculateRepositoryUrl(scRepoBasePath, scRepoPath)
                scRepoPath = trimLeadingAndTrailingSlashes(scRepoPath)
                dirsChanged = trimLeadingAndTrailingSlashes(dirsChanged)
                repoCollection = trimLeadingAndTrailingSlashes(repoCollection)
                if (isSvnRepo) {
                    if (scRepoPath.equals(dirsChanged)) {
                        matchingSourceConfigSet.add(sc)
                        log.debug("Found matching source config from project $sc.project.name with value $finalRepoPath")
                    }
                    else {
                        log.debug("Trigger did not match source config from project $sc.project.name with value $finalRepoPath")
                    }
                }
                else {
                    String dirsChangedCollection = repoCollection + '/' + dirsChanged
                    if (scRepoPath.equals(dirsChangedCollection)) {
                        matchingSourceConfigSet.add(sc)
                        log.debug("Found matching source config from project $sc.project.name with value $finalRepoPath")
                    }
                    else {
                        log.debug("Trigger did not match source config from project $sc.project.name with value $finalRepoPath")
                    }
                }
            }
        }
    }
}
else {
    log.warn("Could not process Svn repo trigger data: $triggerPropertyMap - invalid repo parameters")
}