import java.text.SimpleDateFormat
import java.util.regex.Pattern
import java.util.regex.Matcher
import com.urbancode.anthill3.*

final def ahptool = new AHPTool()
final def dateParser = new DateParser()

final def GMT = TimeZone.getTimeZone("GMT")

final def AHP_DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z");
AHP_DATE_FORMAT.timeZone = GMT

final def SVN_DATE  = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z")
SVN_DATE.timeZone = GMT

final def SVN_DATE_OUT_FORMAT = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.S'Z'")
SVN_DATE_OUT_FORMAT.timeZone = GMT

/**
 * Converts an Ant-style wild-mat string into an equivalent java.util.regex.Pattern
 *   ? is a single char wildcard
 *   * is a wildcard matches any string (but does not cross directories)
 *   ** is a wildcard which matches any string and can cross any number of directories
 */
public def convertToPattern( def antWildMat) {
    // trim leading / character from pattern
    def pattern = antWildMat.startsWith('/') ? antWildMat : '/'+antWildMat

    // deal with special regex-characters that should be interpreted as literals
    '\\.+[]^${}|()'.toCharArray().each{ c ->
        pattern = pattern.replace(''+c, '\\'+c)
    }
    pattern = pattern.replace('?', '.') // ? is a single-char wildcard

    // deal with ant-style wildcards
    StringBuffer result = new StringBuffer()
    def m = (pattern =~ '\\*\\*/|\\*\\*|\\*')
    while (m) {
        def token = m.group()
        def replacement;
        if (token == '**/') {
            replacement = '.*(?<=/)'
        }
        else if (token == '**') {
            replacement = '.*'
        }
        else {
            replacement = '[^/]*'
        }
        m.appendReplacement(result, java.util.regex.Matcher.quoteReplacement(replacement))
    }
    m.appendTail(result)
    return ~(result.toString())
}

private def collectFilters(filters) {
    def result = ["includes": [] as Set, "excludes": [] as Set]
    if (filters) {
        filters.readLines().collect{it.trim()}.findAll{it}.each{filter ->
            def pattern = convertToPattern(filter.substring(1))
            if (filter.startsWith('+')) {
                result.getAt("includes") << pattern
            }
            else if (filter.startsWith('-')) {
                result.getAt("excludes") << pattern
            }
            else {
                // not a valid expression
            }
        }
    }
    return result;
}

//------------------------------------------------------------------------------
// PARSE PROPERTIES
//------------------------------------------------------------------------------

final def props = ahptool.getStepProperties() + ahptool.getJobProperties()

final def workDir = new File('.').canonicalFile
final def out = System.out

final def isSvnRepo     = props['source/repo/isSvnRepo']
final def svnExe        = props['source/repo/commandPath']?: 'svn'
def svnRoot             = props['source/repo/repositoryUrl']
final def svnUsr        = props['source/repo/username']
final def svnPwd        = props['source/repo/password']
final def svnTrunk      = props['source/projectTrunkPath']
final def svnTags       = props['source/projectTagsPath']
final def revision      = props['source/revision'];
final def fromTag       = props['fromTag']
final def startRevision = props['startRevision']
final def endRevision   = props['endRevision']
final def startDate     = dateParser.parseDate(props['startDate']) // TODO fallback to property:workspace.date?
final def endDate       = dateParser.parseDate(props['endDate']) // TODO fallback to property:workspace.date?
final def noCache       = props['cache']

final def sourceUserExcludes  = props['source/excludeUsers']
final def stepUserExcludes    = props['excludeUsers']

def userExcludes = [] as Set
if (sourceUserExcludes) {
    userExcludes.addAll(sourceUserExcludes.readLines().collect{it.trim()}.findAll{it})
}
if (stepUserExcludes) {
    userExcludes.addAll(stepUserExcludes.readLines().collect{it.trim()}.findAll{it})
}

final def sourceFileFilters   = props['source/fileFilters']
final def stepFileFilters     = props['fileFilters']

def includeFiles = [] as Set
def excludeFiles = [] as Set
def sourcePatterns = collectFilters(sourceFileFilters)
includeFiles.addAll(sourcePatterns.includes)
excludeFiles.addAll(sourcePatterns.excludes)
def stepPatterns = collectFilters(stepFileFilters)
includeFiles.addAll(stepPatterns.includes)
excludeFiles.addAll(stepPatterns.excludes)

//determine the checkout path
def checkoutPath = '/'
if (fromTag) {
    checkoutPath += svnTags + '/' + fromTag
}
else {
    checkoutPath += svnTrunk
}
//normalize checkoutPath file separators and trim tail '/' if present
checkoutPath = checkoutPath.replaceAll('/+', '/').replaceAll('/$', '')


if (revision) {
    String msg = "Skipping changelog of revision based checkout for %s:r%s";
    println(String.format(msg, (svnRoot + checkoutPath), revision));
    return;
}


//create a ProcessBuilder with appropriate ticket file in the environment and the working directory
final def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)

//------------------------------------------------------------------------------
//Define convenience methods which depend upon input parameters
//------------------------------------------------------------------------------

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
        proc.out.close() // close stdin
    }
    proc.waitFor()
    if (proc.exitValue()) {
        throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}

/**
 * determine if the given path is a non-excluded path belonging to the desired set
 *
 * @param path the file path to be tested
 */
def pathIsIncluded = { path, repoOffset ->
    boolean included = true;

    def safeCheckoutPath = repoOffset ?: ""

    // check that it matches source being logged
    if (!path.startsWith(safeCheckoutPath)) {
        return false
    }

    // check that it matches source being logged
    if (!path.startsWith(safeCheckoutPath)) {
        return false
    }

    // now check the path (as relative to the basePath) against include/exclude patterns
    def relativePath = path.substring(safeCheckoutPath.length()) // relative path will start with '/'

    if (includeFiles.find{relativePath ==~ it} != null) {
        return true
    }

    if (excludeFiles.find{relativePath ==~ it} != null) {
        return false
    }
    return true;
}

//------------------------------------------------------------------------------
// PREPARE COMMAND LINE
//------------------------------------------------------------------------------
def infoCommand = null
if (noCache) {
    infoCommand = [svnExe, 'info', '--non-interactive', '--no-auth-cache', svnRoot + checkoutPath]
}
else {
    infoCommand = [svnExe, 'info', '--non-interactive', svnRoot + checkoutPath]
}
if (svnUsr) {
    infoCommand << '--username' << svnUsr
    if (svnPwd) {
        infoCommand << '--password' << svnPwd
    }
}

def output = null
runCommand('Getting Repository Info', infoCommand) { Process proc ->
    proc.consumeProcessErrorStream(System.out)
    proc.out.close()
    output = proc.text
    //println output
}
def fullUrl = output.find(~'(?m)^URL: (.*)$') { match, url -> return url }
def root = output.find(~'(?m)^Repository Root: (.*)$') {match, repoOffset -> return repoOffset }

def modulePattern = Pattern.compile('([^/]+)/?$')
def moduleMatcher = modulePattern.matcher(root)

def repoOffset = fullUrl.substring(root.length())
//println "Repository Root: ${repoOffset}"

def changeSets = []

def command = null
if (noCache) {
    command = [svnExe, 'log', '--non-interactive', '--no-auth-cache', '--verbose', '--xml']
}
else {
    command = [svnExe, 'log', '--non-interactive', '--verbose', '--xml']
}

if (svnUsr) {
    command << '--username' << svnUsr
    if (svnPwd) {
        command << '--password' << svnPwd
    }
}

if (startRevision || startDate) {
    command << '--revision'
    String revArg = startRevision ?: ('{'+SVN_DATE.format(startDate)+'}')
    if (endRevision){
        revArg += ':'+endRevision
    }
    else if (endDate) {
        revArg += ':'+'{'+SVN_DATE.format(endDate)+'}'
    }

    command << revArg
}

command << (svnRoot + checkoutPath)


def svnLog = null
runCommand('Getting Changelog', command) { proc ->
    proc.consumeProcessErrorStream(System.out)
    proc.out.close()
    svnLog = proc.text
    println svnLog
}

def svnModule = null
new groovy.util.XmlSlurper().parseText(svnLog).logentry.each{ svnchangeset ->
    if (moduleMatcher) {
        svnModule = moduleMatcher.group(1)
    }
    def svnChangesetDateText = svnchangeset.date[0].text()
    svnChangesetDateText = svnChangesetDateText.replaceFirst('[0-9]{3}Z$', 'Z')
    def svnDate = SVN_DATE_OUT_FORMAT.parse(svnChangesetDateText) // 2009-05-06T20:12:41.284343Z

    def changeSet = [id:null, user:null, message:null, date:null, fileSet:[]]
    changeSet.id      = svnchangeset.@revision.text()
    changeSet.user    = svnchangeset.author[0].text()
    changeSet.message = svnchangeset.msg[0].text()
    changeSet.date    = AHP_DATE_FORMAT.format(svnDate)

    boolean hasAllowedAuthor = !userExcludes || !userExcludes.contains(changeSet.user)
    boolean hasAllowedPaths = false

    svnchangeset.paths[0].path.each{ path ->
        def operation = path.@action.text()
        def filename = path.text()
        changeSet.fileSet << [type: operation, rev: changeSet.id, path: filename]
        hasAllowedPaths = hasAllowedPaths || pathIsIncluded(filename, repoOffset)
    }

    boolean allowedDate = true
    if (startDate && endDate) {
        allowedDate = svnDate.after(startDate) && svnDate.before(endDate)
    }

    if (hasAllowedAuthor && hasAllowedPaths && allowedDate) {
        changeSets << changeSet
    }
    else {
        def message = new StringBuilder("Changeset ${changeSet.id} skipped because ")
        if (!hasAllowedAuthor) {
            message << "it has excluded author ${changeSet.user}"
        }
        if (!hasAllowedPaths) {
            if (!hasAllowedAuthor) {
                message << " and "
            }
            message << "it contains only excluded file paths (${changeSet.fileSet.collect{it.path}})"
        }
        if (!allowedDate) {
            if (!hasAllowedAuthor || !hasAllowedPaths) {
                message << " and "
            }
            message << "it has an invalid date (" + changeSet.date + ") for the given date range"
        }
        println message
    }
}

//------------------------------------------------------------------------------
// Construct XML message and upload to AHP
//------------------------------------------------------------------------------

if (changeSets) {
    def xml = groovy.xml.DOMBuilder.newInstance().'change-log'{
        changeSets.each{ changeSet ->
            'change-set'(){
                'id'(changeSet.id)
                if (svnModule) {
                    'module'(svnModule)
                }
                'user'(changeSet.user)
                'date'(changeSet.date)
                if (changeSet.message) {
                    'comment'(changeSet.message)
                }
                if (fromTag) {
                    'branch'(svnTags + '/' + fromTag)
                }
                'repository-type'('svn')
                'repository-id'(System.getenv()['AH_REPOSITORY_ID'])
                'file-set'{
                    changeSet.fileSet.each{change ->
                        'file'('change-type':change.type, 'revision-number':change.rev, change.path)
                    }
                }
            }
        }
    }

    ahptool.addChangeSets(xml)
}
else {
    println "No changes detected"
}
