import com.urbancode.anthill3.*

final def workDir = new File('.').canonicalFile
final def out = System.out
final def env = System.getenv()

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = new AHPTool()
final def dateParser = new DateParser()

final def DATE_READ_1 = new java.text.SimpleDateFormat("E MMM dd HH:mm:ss zzz yyyy") // java.util.Date.toString()
final def DATE_READ_2 = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss z") // ISO-like format
final def SVN_DATE = new java.text.SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z")
SVN_DATE.timeZone = TimeZone.getTimeZone("GMT")

//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS
//------------------------------------------------------------------------------
def props = ahptool.getStepProperties()

final def svnExe       = props['source/repo/commandPath']?: 'svn'
final def svnRoot      = props['source/repo/repositoryUrl']
final def svnUsr       = props['source/repo/username']
final def svnPwd       = props['source/repo/password']
final def svnTrunk     = props['source/projectTrunkPath']
final def svnTags      = props['source/projectTagsPath']
final def dir          = new File(workDir, props['source/dirOffset'] ?: '.').canonicalFile // get checkout directory

final def fromTag      = props['fromTag']
final def revision     = props['revision'] ?: props['source/revision']
final def date         = dateParser.parseDate(props['date'])
final def depth        = props['depth']
final def noCache      = props['cache']

final boolean doExport = props['export']?.toBoolean()
final boolean clean    = props['cleanWorkspace']?.toBoolean()

final def commandHelper = new CommandHelper(workDir);

//------------------------------------------------------------------------------
// PREPARE WORKING DIRECTORY
//------------------------------------------------------------------------------

// clean && create the working directory
if (clean && dir.isDirectory()) {
    new AntBuilder().delete(includeemptydirs:'true'){fileset(dir: workDir.path, includes:'**/*', defaultexcludes:'false')}
}
dir.mkdirs()
if (!dir.exists()) {
    throw new Exception("Could not create working directory ${workDir}")
}
else if (dir.isFile()) {
    throw new Exception("Specified working directory is a file!! ${workDir}")
}

//------------------------------------------------------------------------------
// PREPARE COMMAND LINE
//------------------------------------------------------------------------------

// determine the checkout path
def checkoutPath = '/'
if (fromTag) {
    checkoutPath += svnTags + '/' + fromTag
}
else {
    checkoutPath += svnTrunk
}

// normalize checkoutPath file separators and trim tail '/' if present
checkoutPath = checkoutPath.replaceAll('/+', '/').replaceAll('/$', '')

def command = null
if (noCache) {
    command = [svnExe, '--non-interactive', '--no-auth-cache']
}
else {
    command = [svnExe, '--non-interactive']
}

if (depth && depth != "infinity") {
    command << '--depth' << depth
}
if (doExport) {
    command << 'export' << '--force' // need force because target directory does/may exist
}
else {
    command << 'checkout'
}

if (svnUsr) {
    command << '--username' << svnUsr
    if (svnPwd) {
        command << '--password' << svnPwd
    }
}

if (revision) {
    command << '--revision' << revision
}
else if (date) {
    command << '--revision' << ('{'+SVN_DATE.format(date)+'}')
}

command << svnRoot + checkoutPath

command << dir.path // check out to workDir

//------------------------------------------------------------------------------
// EXECUTE
//------------------------------------------------------------------------------

// prepare process and set environment: MUST cast command to String[] !!!
commandHelper.runCommand('Subversion Checkout', command);

// run svn info on the workspace and set the revision as a property on the build life
def infoCommand = null
if (noCache) {
    infoCommand = [svnExe, 'info', '--non-interactive', '--no-auth-cache']
}
else {
    infoCommand = [svnExe, 'info', '--non-interactive']
}

if (svnUsr) {
    infoCommand << '--username' << svnUsr
    if (svnPwd) {
        infoCommand << '--password' << svnPwd
    }
}

if (revision) {
    infoCommand << '--revision' << revision
}
else if (date) {
    infoCommand << '--revision' << ('{'+SVN_DATE.format(date)+'}')
}

infoCommand << svnRoot + checkoutPath

def output = null
commandHelper.runCommand('Getting Repository Info', infoCommand) { Process proc ->
    proc.out.close() // close stdin
    def t = proc.consumeProcessErrorStream(System.out) // forward stderr
    output = proc.text
    t.join(); // wait for async err thread to complete
}

def currentRevision = output.find(~'(?m)^Revision: (.*)$') {match, rev -> return rev }
def changedRevision = output.find(~'(?m)^Last Changed Rev: (.*)$') {match, rev -> return rev}
def currentBuildLifeProps = ahptool.getBuildLifeProperties();
def newBuildLifeProps = ['svn-revision' : currentRevision, 'svn-changed-revision' : changedRevision] + currentBuildLifeProps
ahptool.setBuildLifeProperties(newBuildLifeProps)