import java.text.SimpleDateFormat
import com.urbancode.anthill3.*

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptoolExe = isWindows ? 'ahptool.cmd' : 'ahptool'
final def ahptool = new AHPTool()
final def dateParser = new DateParser()

public final def DATE_READ_1 = new SimpleDateFormat("E MMM dd HH:mm:ss zzz yyyy") // java.util.Date.toString()
public final def DATE_READ_2 = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z") // ISO-like format
public final def SVN_DATE  = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z")
SVN_DATE.timeZone = TimeZone.getTimeZone("GMT")

// Svn xml date format always in GMT and of the form 2008-12-22T13:09:32.011244Z
public final def SVN_XML_DATE = new SimpleDateFormat("yyyy-MM-dd'T'hh:mm:ss.S'Z'")
SVN_XML_DATE.timeZone = TimeZone.getTimeZone("GMT")

/**
 * Converts an Ant-style wild-mat string into an equivalent java.util.regex.Pattern
 *   ? is a single char wildcard
 *   * is a wildcard matches any string (but does not cross directories)
 *   ** is a wildcard which matches any string and can cross any number of directories
 */
public def convertToPattern( def antWildMat) {
    // trim leading / character from pattern
    def pattern = antWildMat.startsWith('/') ? antWildMat : '/'+antWildMat

    // deal with special regex-characters that should be interpreted as literals
    '\\.+[]^${}|()'.toCharArray().each{ c ->
        pattern = pattern.replace(''+c, '\\'+c)
    }
    pattern = pattern.replace('?', '.') // ? is a single-char wildcard

    // deal with ant-style wildcards
    StringBuffer result = new StringBuffer()
    def m = (pattern =~ '\\*\\*/|\\*\\*|\\*')
    while (m) {
        def token = m.group()
        def replacement;
        if (token == '**/') {
            replacement = '.*(?<=/)'
        }
        else if (token == '**') {
            replacement = '.*'
        }
        else {
            replacement = '[^/]*'
        }
        m.appendReplacement(result, java.util.regex.Matcher.quoteReplacement(replacement))
    }
    m.appendTail(result)
    return ~(result.toString())
}

//------------------------------------------------------------------------------
// PARSE PROPERTIES
//------------------------------------------------------------------------------

final def workDir = new File('.').canonicalFile
final def out = System.out
final def env = System.getenv()

// use ahptool to read all properties associated with this step
def props = ahptool.getStepProperties()

final def isSvnRepo     = props['source/repo/isSvnRepo']
final def svnExe        = props['source/repo/commandPath']?: 'svn'
def svnRoot             = props['source/repo/repositoryUrl'] // not final so we can prepend http:// if no protocol is specified
final def svnUsr        = props['source/repo/username']
final def svnPwd        = props['source/repo/password']
final def svnTrunk      = props['source/projectTrunkPath']
final def svnTags       = props['source/projectTagsPath']
final def revision      = props['source/revision'];

final def fromTag       = props['fromTag']

final def startRevision = props['startRevision']
final def endRevision   = props['endRevision']
final def startDate     = dateParser.parseDate(props['startDate'])
final def endDate       = dateParser.parseDate(props['endDate'])
final def noCache       = props['cache']

def collectionDirectory = ""

def userExcludes = props['excludeUsers']
if (userExcludes) {
    userExcludes = userExclude.readLines().collect{it.trim()}.findAll{it}
}

def fileFilters = props['fileFilters']
def includeFiles = []
def excludeFiles = []

if (fileFilters) {
    fileFilters.readLines().collect{it.trim()}.findAll{it}.each{filter ->
        def pattern = convertToPattern(filter.substring(1))
        if (filter.startsWith('+')) {
            includeFiles << pattern
        }
        else if (filter.startsWith('-')) {
            excludeFiles << pattern
        }
        else {
            // not a valid expression
        }
    }
}

// determine the checkout path
def checkoutPath = '/'
if (fromTag) {
    checkoutPath += svnTags + '/' + fromTag
}
else {
    checkoutPath += svnTrunk
}

// normalize checkoutPath file separators and trim tail '/' if present
checkoutPath = checkoutPath.replaceAll('/+', '/').replaceAll('/$', '')

/**
 * determine if the given path is a non-excluded path belonging to the desired trunk/banch/tag
 */
def pathIsIncluded = { path, repoOffset ->
    boolean included = true;

    def safeCheckoutPath = repoOffset ?: ""

    // check that it matches source being logged
    if (!path.startsWith(safeCheckoutPath)) {
        return false
    }

    // now check the path (as relative to the basePath) against include/exclude patterns
    def relativePath = path.substring(safeCheckoutPath.length()) // relative path will start with '/'

    for (include in includeFiles) {
        if (relativePath ==~ include) {
            return true
        }
    }

    for (exclude in excludeFiles) {
        if (relativePath ==~ exclude) {
            return false
        }
    }
    return true;
}

//
// Validate
//

if (!svnRoot) {
    throw new Exception("An svn Repository Url must be specified")
}
else if (!svnRoot.startsWith('http') && !svnRoot.startsWith('svn:')) {
    // should we prepend 'http://' or throw exception ????
    //if (!(svnRoot =~ /^(http|https|svn):[/][/]/)) svnRoot = 'http://'+svnRoot
}

if (startRevision && startDate) {
    throw new Exception('Both a startRevision and startDate can not be specified at the same time')
}
if (startRevision == null && startDate != null) {
    startRevision = '{'+SVN_DATE.format(startDate)+'}'
}

// TODO should always be an endDate (or revision)?
if (endRevision && endDate) {
    throw new Exception('Both a endRevision and endDate can not be specified at the same time')
}
if (endRevision == null && endDate != null) {
    endRevision = '{'+SVN_DATE.format(endDate)+'}'
}

if (fromTag && !svnTags) {
    // ????
}

if (revision) {
    String msg = "Skipping changelog of revision based checkout for %s:r%s";
    println(String.format(msg, (svnRoot + checkoutPath), revision));
    return;
}


//------------------------------------------------------------------------------
// PREPARE COMMAND LINE
//------------------------------------------------------------------------------
//  svn log --revision '{2008-12-22 08:09:39 -0500}:{2008-12-15 09:47:55 -0500}' --verbose --xml path
//     note: can not use the "--limit X" flag because the latest change could be from an ignored author or contain only excluded files
//

def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)

/**
 * A convenience method for running commands and optionally parsing the stdout of that command.
 * If closure is non-null, the closure will be passed the resultant {@link Process} and is expected to deal with all IO.
 * Otherwise, the process' stdOut and stdErr are forwarded to this scripts 'stdOut' and stdIn is untouched.
 *
 * @param message an optional message to print prior to the commandline
 * @param command the command to be used as a String[]
 * @param closure an optional closure to deal with Process IO
 */
final def runCommand = {def message, def command, def closure ->
    pb.command(command as String[])
    println()
    if (message) {
        println(message)
    }
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    if (closure) {
        closure(proc)
    }
    else {
        proc.consumeProcessOutput(out, out) // forward stdout and stderr
        proc.getOutputStream().close() // close stdin
    }
    proc.waitFor()
    if (proc.exitValue()) {
        throw new Exception("Command failed with exit code: "+proc.exitValue())
    }
}


def command = null
if (noCache) {
    command = [svnExe, 'log', '--non-interactive', '--no-auth-cache', '--verbose', '--xml']
}
else {
    command = [svnExe, 'log', '--non-interactive', '--verbose', '--xml']
}

if (svnUsr) {
    command << '--username' << svnUsr
    if (svnPwd) {
        command << '--password' << svnPwd
    }
}

// order range as  end:start so that limiting will give the one closest to end-date
if (startRevision || endRevision) {
    command << '--revision'
    def revArg = new StringBuilder()

    if (endRevision) {
        revArg += endRevision
    }
    if (startRevision) {
        if (revArg.length()) {
            revArg += ':'
        }
        revArg += startRevision
    }
    command << revArg
}

command << (svnRoot + checkoutPath)

//------------------------------------------------------------------------------
// BUILD PROCESSES AND EXECUTE
//------------------------------------------------------------------------------
def infoCommand = null
if (noCache) {
    infoCommand = [svnExe, 'info', '--non-interactive', '--no-auth-cache', svnRoot + checkoutPath]
}
else {
    infoCommand = [svnExe, 'info', '--non-interactive', svnRoot + checkoutPath]
}

if (svnUsr) {
    infoCommand << '--username' << svnUsr
    if (svnPwd) {
        infoCommand << '--password' << svnPwd
    }
}
def output = null
runCommand('Getting Repository Info', infoCommand) { Process proc ->
    proc.consumeProcessErrorStream(System.out)
    proc.out.close()
    output = proc.text
    //println output
}
def fullUrl = output.find(~'(?m)^URL: (.*)$') { match, url -> return url }
def root = output.find(~'(?m)^Repository Root: (.*)$') {match, repoOffset -> return repoOffset }

def repoOffset = fullUrl.substring(root.length())
//println "Repository Root: ${repoOffset}"


def logXmlResult = '' // contain the output of logProc

runCommand('Getting Changelog', command) { proc ->
    proc.out.close() // close svn stdIn
    proc.consumeProcessErrorStream(System.out) // forward svn stdErr to our stdout
    logXmlResult = proc.text; // capture svn stdOut (synchronous)
    println(logXmlResult) // echo the xml
}

def doc = new XmlSlurper().parseText(logXmlResult.toString())
def logEntries = doc.logentry
if (userExcludes) {
    // filter out excluded authors
    logEntries = logEntries.findAll{!userExcludes.contains(it.author[0].text())}
}

// find entries which reference a non-excluded path (and which belong to this trunk/branch/tag)
logEntries = logEntries.findAll{
    def paths = it.paths[0].path.collect{it.text()}
    return (boolean)(paths.find{path->pathIsIncluded(path, repoOffset)})
}

// locate the latest date (expressed as milliseconds since epoch)
Long latestDate = logEntries.collect{SVN_XML_DATE.parse(it.date[0].text()).time}.max()

//------------------------------------------------------------------------------
// UPLOAD LATEST CHANGE DATE
//------------------------------------------------------------------------------

if (latestDate == null) {
    println('No changes detected')
}
else {
    def uploadChangeDateCmd = [ahptoolExe, 'setStepProperties', '-']

    def propXml = groovy.xml.DOMBuilder.newInstance().'properties'{
        property('name':'latest-change', latestDate)
    }

    runCommand('Uploading time of latest change ', uploadChangeDateCmd) { proc ->
        proc.consumeProcessOutput(out, out) // forward process stdout and stderr to our stdout
        proc.withWriter{ it << propXml }
    }
}