import java.text.SimpleDateFormat

final def workDir = new File('.').canonicalFile
final def out = System.out
final def env = System.getenv()

final def isWindows = (System.getProperty('os.name') =~ /(?i)windows/).find()
final def ahptool = isWindows ? 'ahptool.cmd' : 'ahptool'

final def SVN_DATE = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z")
SVN_DATE.timeZone = TimeZone.getTimeZone("GMT")

final def DATE_FORMATS = [
    new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z"),    // SVN
    new SimpleDateFormat('yyyy-MM-dd HH:mm:ss z'),      // ISO-like
    new SimpleDateFormat('E MMM dd HH:mm:ss zzz yyyy'), // java.util.Date.toString()
    new SimpleDateFormat('yyyy-MM-dd hh:mm:ss.S a z') // AHP
]

DATE_FORMATS.each{ dateFormat ->
    dateFormat.timeZone = TimeZone.getTimeZone("GMT")
}

/**
 * Try a bunch of methods/formats to parse the given value into a Date object
 *
 * @param value the value to parse
 * @return the corresponding date
 * @throws java.text.ParseException if the date can not be parsed
 */
final def parseDate = { value ->
    if (!value) {
        return null
    }
    if (value.isLong()) {
        return new Date(value.toLong())
    }
    else {
        def finalException = null
        for (dateFormat in DATE_FORMATS) {
            try {
                return dateFormat.parse(value)
            }
            catch (java.text.ParseException e) {
                finalException = e
            }
        }
        if (finalException) {
            throw finalException
        }
    }
}

//------------------------------------------------------------------------------
// GET ALL INPUT PARAMETERS AND VALIDATE
//------------------------------------------------------------------------------

def props = [:]
['Job', 'Step'].each { scope ->
  def ahptoolProc = (ahptool + " get${scope}Properties").execute()
  ahptoolProc.consumeProcessErrorStream(System.out);
  def propertiesXml = ahptoolProc.text
  ahptoolProc.waitFor()
  if (ahptoolProc.exitValue()) {
      throw new Exception("Failed to get step properties from ahptool: " + propertiesXml)
  }
  new XmlSlurper().parseText(propertiesXml).property.each{ propElem ->
    props[propElem.@name.text()] = propElem.text()
  }
}

final def svnExe       = props['source/repo/commandPath']?: 'svn'
final def svnRoot      = props['source/repo/repositoryUrl']
final def svnUsr       = props['source/repo/username']
final def svnPwd       = props['source/repo/password']
final def svnTrunk     = props['source/projectTrunkPath']
final def svnBranches  = props['source/projectBranchesPath']
final def svnTags      = props['source/projectTagsPath']

final def fromBranch   = props['fromBranch']
final def fromTag      = props['fromTag']
final def revision     = props['revision'] ?: props['source/revision']
final def date         = parseDate(props['date'] ?: props['workspace.date'])

final def createBranch = props['createBranch']
final def createTag    = props['createTag']
final def message      = props['commitMessage']
final def noCache      = props['cache']

final def pb = new ProcessBuilder(['echo'] as String[]).directory(workDir)

//------------------------------------------------------------------------------
// PREPARE COMMAND LINE
//------------------------------------------------------------------------------

    def command = null
    if (noCache) {
        command = [svnExe, '--non-interactive', '--no-auth-cache', 'copy']
    }
    else {
        command = [svnExe, '--non-interactive', 'copy']
    }

    if (svnUsr) {
        command << '--username' << svnUsr
        if (svnPwd) {
            command << '--password' << svnPwd
        }
    }

    command << '--message' << message

    // label from sourceurl
    if (revision) {
        command << '--revision' << revision
    }
    else if (date) {
        command << '--revision' << ('{'+SVN_DATE.format(date)+'}')
    }

    if (fromTag) {
        command.add(svnRoot+'/'+svnTags+'/'+fromTag)
    }
    else {
        command.add(svnRoot+'/'+svnTrunk)
    }

    // destination url

    if (createBranch) {
        command.add(svnRoot+'/'+svnBranches+'/'+createBranch)
    }
    else if (createTag) {
        command.add(svnRoot+'/'+svnTags+'/'+createTag)
    }
    else {
        throw new IllegalStateException() // shouldn't happen
    }


//------------------------------------------------------------------------------
// EXECUTE
//------------------------------------------------------------------------------
    pb.command(command as String[])
    println("command: ${pb.command().join(' ')}")
    def proc = pb.start()
    proc.consumeProcessOutput(out, out)
    proc.waitFor()
    if (proc.exitValue()) {
      throw new Exception("Command failed with exit code: "+proc.exitValue())
    }